/*-------------------------------------------------------------------------
 *
 * pg_audit_users.c
 *		Track connection and activity of users
 *
 * Copyright (c) 2014, Pierre Ducroquet <pinaraf@pinaraf.info>
 *
 * Loosely based on pg_stat_statements
 * Copyright (c) 2008-2012, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *	  contrib/pg_audit_users/pg_audit_users.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include "access/hash.h"
#include "miscadmin.h"
#include "funcapi.h"
#include "storage/ipc.h"
#include "utils/builtins.h"
#include "utils/timestamp.h"


PG_MODULE_MAGIC;


#define PGAU_DUMP_FILE "global/pg_audit_users.stat"
#define PGAU_MAX_USERS 100
static const uint32 PGAU_FILE_HEADER = 0x20141219;


// Hashkey = Oid of user
// Entry = {Oid of user, pg_tm of last query}
static HTAB *pgau_hash = NULL;

typedef struct pgauEntry
{
	Oid			user;
	TimestampTz 		lastquery;
} pgauEntry;

/* Saved hook values in case of unload */
static shmem_startup_hook_type prev_shmem_startup_hook = NULL;
static ExecutorStart_hook_type prev_ExecutorStart_hook = NULL;

/*---- Function declarations ----*/

void		_PG_init(void);
void		_PG_fini(void);

Datum           pg_audit_users(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(pg_audit_users);

static void pgau_shmem_startup(void);
static void pgau_shmem_shutdown(int code, Datum arg);
static void pgau_ExecutorStart_hook(QueryDesc *queryDesc, int eflags);

/*
 * Module load callback
 */
void
_PG_init(void)
{
	prev_ExecutorStart_hook = ExecutorStart_hook;
	ExecutorStart_hook = pgau_ExecutorStart_hook;
        prev_shmem_startup_hook = shmem_startup_hook;
        shmem_startup_hook = pgau_shmem_startup;
}


void
_PG_fini(void)
{
	// Restore hooks
	ExecutorStart_hook = prev_ExecutorStart_hook;
        shmem_startup_hook = prev_shmem_startup_hook;
}


static void
pgau_ExecutorStart_hook(QueryDesc *queryDesc, int eflags)
{
	pgauEntry *entry = 0;
	HASH_SEQ_STATUS hash_seq;

	if (prev_ExecutorStart_hook)
		(*prev_ExecutorStart_hook) (queryDesc, eflags);
        else
                standard_ExecutorStart(queryDesc, eflags);

	if (!pgau_hash)
		return;

	Oid uid = GetUserId();
	bool found = false;
	
	entry = (pgauEntry *)hash_search(pgau_hash, &uid, HASH_ENTER, &found);
	entry->lastquery = GetCurrentTimestamp();
}

#define PG_AUDIT_USERS_COLUNS_V1 2

/*
 * Retrieve audit data,
 */
Datum
pg_audit_users(PG_FUNCTION_ARGS)
{
        ReturnSetInfo *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
        TupleDesc       tupdesc;
        Tuplestorestate *tupstore;
        MemoryContext per_query_ctx;
        MemoryContext oldcontext;
        HASH_SEQ_STATUS hash_seq;
        pgauEntry  *entry;

        if (!pgau_hash)
                ereport(ERROR,
                                (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
                                 errmsg("pg_audit_users must be loaded via shared_preload_libraries")));

        /* check to see if caller supports us returning a tuplestore */
        if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
                ereport(ERROR,
                                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                                 errmsg("set-valued function called in context that cannot accept a set")));
        if (!(rsinfo->allowedModes & SFRM_Materialize))
                ereport(ERROR,
                                (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                                 errmsg("materialize mode required, but it is not " \
                                                "allowed in this context")));

        /* Build a tuple descriptor for our result type */
        if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
                elog(ERROR, "return type must be a row type");

        per_query_ctx = rsinfo->econtext->ecxt_per_query_memory;
        oldcontext = MemoryContextSwitchTo(per_query_ctx);

        tupstore = tuplestore_begin_heap(true, false, work_mem);
        rsinfo->returnMode = SFRM_Materialize;
        rsinfo->setResult = tupstore;
        rsinfo->setDesc = tupdesc;

        MemoryContextSwitchTo(oldcontext);

        //LWLockAcquire(pgss->lock, LW_SHARED);

        hash_seq_init(&hash_seq, pgau_hash);
        while ((entry = hash_seq_search(&hash_seq)) != NULL)
        {
                Datum           values[PG_AUDIT_USERS_COLUNS_V1];
                bool            nulls[PG_AUDIT_USERS_COLUNS_V1];
                int                     i = 0;

                memset(values, 0, sizeof(values));
                memset(nulls, 0, sizeof(nulls));

                values[i++] = ObjectIdGetDatum(entry->user);
		values[i++] = TimestampTzGetDatum(entry->lastquery);

                tuplestore_putvalues(tupstore, tupdesc, values, nulls);
        }

        //LWLockRelease(pgss->lock);

        /* clean up and return the tuplestore */
        tuplestore_donestoring(tupstore);

        return (Datum) 0;
}


static void 
pgau_shmem_shutdown(int code, Datum arg)
{
	// Clean up the mess
}

static void
pgau_shmem_startup(void)
{
	//bool		found;
	HASHCTL		info;
	//FILE	   *file;
	//uint32		header;
	//int32		num;
	//int32		i;
	//char	   *buffer = NULL;

	if (prev_shmem_startup_hook)
		prev_shmem_startup_hook();

	/* reset in case this is a restart within the postmaster */
	pgau_hash = NULL;

	/*
	 * Create or attach to the shared memory state, including hash table
	 */
	LWLockAcquire(AddinShmemInitLock, LW_EXCLUSIVE);

	memset(&info, 0, sizeof(info));
	info.keysize = sizeof(Oid);
	info.entrysize = sizeof(pgauEntry);
	info.hash = oid_hash;
	pgau_hash = ShmemInitHash("pg_audit_users hash",
							  PGAU_MAX_USERS, PGAU_MAX_USERS,
							  &info,
							  HASH_ELEM | HASH_FUNCTION);
	if (!pgau_hash)
	{
		elog(FATAL, "Could not get my hash !");
	}
	LWLockRelease(AddinShmemInitLock);

	/*
	 * If we're in the postmaster (or a standalone backend...), set up a shmem
	 * exit hook to dump the statistics to disk.
	 */
	if (!IsUnderPostmaster)
		on_shmem_exit(pgau_shmem_shutdown, (Datum) 0);

#if 0
	/*
	 * Attempt to load old statistics from the dump file, if this is the first
	 * time through and we weren't told not to.
	 */
	if (found || !pgss_save)
		return;

	/*
	 * Note: we don't bother with locks here, because there should be no other
	 * processes running when this code is reached.
	 */
	file = AllocateFile(PGSS_DUMP_FILE, PG_BINARY_R);
	if (file == NULL)
	{
		if (errno == ENOENT)
			return;				/* ignore not-found error */
		goto error;
	}

	buffer_size = query_size;
	buffer = (char *) palloc(buffer_size);

	if (fread(&header, sizeof(uint32), 1, file) != 1 ||
		header != PGSS_FILE_HEADER ||
		fread(&num, sizeof(int32), 1, file) != 1)
		goto error;

	for (i = 0; i < num; i++)
	{
		pgssEntry	temp;
		pgssEntry  *entry;

		if (fread(&temp, offsetof(pgssEntry, mutex), 1, file) != 1)
			goto error;

		/* Encoding is the only field we can easily sanity-check */
		if (!PG_VALID_BE_ENCODING(temp.key.encoding))
			goto error;

		/* Previous incarnation might have had a larger query_size */
		if (temp.query_len >= buffer_size)
		{
			buffer = (char *) repalloc(buffer, temp.query_len + 1);
			buffer_size = temp.query_len + 1;
		}

		if (fread(buffer, 1, temp.query_len, file) != temp.query_len)
			goto error;
		buffer[temp.query_len] = '\0';

		/* Skip loading "sticky" entries */
		if (temp.counters.calls == 0)
			continue;

		/* Clip to available length if needed */
		if (temp.query_len >= query_size)
			temp.query_len = pg_encoding_mbcliplen(temp.key.encoding,
												   buffer,
												   temp.query_len,
												   query_size - 1);

		/* make the hashtable entry (discards old entries if too many) */
		entry = entry_alloc(&temp.key, buffer, temp.query_len, false);

		/* copy in the actual stats */
		entry->counters = temp.counters;
	}

	pfree(buffer);
	FreeFile(file);

	/*
	 * Remove the file so it's not included in backups/replication slaves,
	 * etc. A new file will be written on next shutdown.
	 */
	unlink(PGSS_DUMP_FILE);

	return;

error:
	ereport(LOG,
			(errcode_for_file_access(),
			 errmsg("could not read pg_stat_statement file \"%s\": %m",
					PGSS_DUMP_FILE)));
	if (buffer)
		pfree(buffer);
	if (file)
		FreeFile(file);
	/* If possible, throw away the bogus file; ignore any error */
	unlink(PGSS_DUMP_FILE);
#endif
}


